/* Mirror directrory structure to another host                   */
/*                                                               */
/* Usage: mirror [-v][-d][-l][-e][-c] local_dir host user passwd remote_dir */ 
/*                                                               */
/*    -v verbose                                                 */
/*    -d debug                                                   */
/*    -c do a cmp before moving into final position             */
/*    -l only list, don't copy                                   */
/*    -e attempt to extract readme files (requires dos & tar)    */
/*                                                               */
/* Performs a mirror of the remote directory to the local using  */
/* the ftp protocol.  Sub directories are also mirrored.  Files  */
/* and directories are deleted on the local machine when required*/
/* Symbolic links are also mirrored                              */
/*                                                               */
/* The "readme" file of .exe's and .tar's are extracted on copy  */
/* if possible                                                   */
/*                                                               */
/* Mirroring is based on timestamp and file size info            */
/*                                                               */
/* Ftp traffic and disk access is minimised                      */
/*                                                               */
/* If password is "-", then password is read on stdin            */
/*                                                               */
/* ideas for future enhancements :-                              */
/*                                                               */
/*   Read full perms of remote machine to mirror perms as well   */
/*   More efficient backup by moving files instead of deleting   */
/*     then check the list of "moved" files before every get     */
/*     (really del the temp directory before exiting)            */
/*   Delete local hard linked files as well                      */
/*   Ability to mirror from more than one machine into one dir   */
/*   Improve the code structure !!!                              */
/*   Clean up temp files on abnormal quit                        */
/*   Mirror on the basis of the remote ls-lR.Z file              */
/*   Send logs to syslog instead of stderr                       */
/*                                                               */
/*                                                               */
/*   Peter Lord (plord@novell.co.uk)                             */
/*                                                               */
/*   (C) Novell 1995                                             */
/*                                                               */
/*   Comments/suggestions/re-writes etc etc more than welcome!   */
/*                                                               */
/*   Note that on 1.1.1, /dev/tcp doesn't seem to be closed      */
/*   Causing out of open files if mirroring a large tree ...     */
/*                                                               */
/*                                                               */
/*    Based on ideas in uftp & uses libftp                       */
/*                                                               */
/*    Both uftp & libftp was written by Oleg Orel                */
/*    Department of Electronics and automatisation.              */
/*    Institute for High Energy Physics                          */
/*    Protvino, Russia                                           */
/*    E-mail: orel@oea.ihep.su, orel@dxcern.cern.ch              */

static char id[] = "Mirror version 0.6, (c) Novell UK 1995.";


#include <stdio.h>
#include <sys/types.h>
#include <sys/statvfs.h>
#include <sys/time.h>
#include <sys/dir.h>
#include <sys/stat.h>
#include <syslog.h>
#include <dirent.h>
#include <string.h>
#include <varargs.h>
#include <utime.h>
#include <unistd.h>
#include <setjmp.h>
#include <ulimit.h>
#include <signal.h>
#include <FtpLibrary.h>

FTP *ftp;
int verbose, debug, list, extract, cmp;
char host[MAXNAMLEN], user[MAXNAMLEN], pass[MAXNAMLEN], pwd[MAXNAMLEN];
long maxfilesize;

jmp_buf getstack;

/* #define DEBUGMALLOC */
/* #define DEBUGTEMP */
#define LOGINTRIES 10
#define LOGINDELAY 600

/* keep 10Mb free ... */
#define FREESPACE 10*1024*1024

#define USAGE "Usage: mirror [-v][-d][-l][-e][-c] local_dir host user passwd remote_dir\n\n	-v verbose\n	-d debug\n	-l only list, don't copy\n	-e attempt to extract file.txt (needs merge & tar)\n	-c perform a cmp before moving in position\n\n%s\n",id

void mylog (format, va_alist) const char *format; va_dcl
{
  if (verbose)
  {
    /* log handler */
    va_list ap;
    va_start(ap);
    fprintf (stderr, "mirror: ");
    vfprintf(stderr, format, ap);
    va_end(ap);
  }
}

void myerror (format, va_alist) const char *format; va_dcl
{
  /* non-quit error handler */
  va_list ap;
  va_start(ap);
  fprintf (stderr, "mirror: Error: ");
  vfprintf(stderr, format, ap);
  va_end(ap);
}

void myquit (format, va_alist) const char *format; va_dcl
{
  /* quit handler */
  va_list ap;
  va_start(ap);
  fprintf (stderr, "mirror: Error: ");
  vfprintf(stderr, format, ap);
  va_end(ap);
  if (ftp > (FTP *)0)
    FtpBye(ftp);
  exit(1);
}

mycopy (char *name, time_t mtime, off_t size)
{
  /* copy a file and "touch" it */
  int a;
  struct utimbuf times;
  struct statvfs info;
  char basename[MAXNAMLEN], command[MAXNAMLEN], thisdirname[MAXNAMLEN];

  /* first check to see if we have some free space */

  strcpy (basename, name);
  *(strrchr(basename, '/')) = 0;
  if (statvfs (basename, &info) < 0)
    mylog ("Cannot statvfs %s\n", basename);
  else
  {
    if (((info.f_bavail*info.f_frsize)-size)  < FREESPACE )
      myquit ("Out of free space on %s, quiting\n", basename);
  }

  if (size>maxfilesize)
    mylog ("Cannot mirror %s, too large\n", name);
  else
  {

    /* hmmmm ... FtpGet seems to not release file handles on permission denied */
    if (cmp == 0)
      a = FtpGet (ftp, name, name);
    else
    {
      a = FtpGet (ftp, name, ".cmpdiff");
      sprintf (command, "/usr/bin/cmp .cmpdiff %s >/dev/null 2>/dev/null", name);
      if (system(command) != 0)
        rename (".cmpdiff", name);
      else
        unlink (".cmpdiff");
    }

    if (a < 0)
    {
      if (a == -550)
        myerror ("Cannot mirror %s, permission denied\n", name);
      else
        myerror ("Cannot mirror %s\n", name);
      unlink (name);
    }
    else
    {
      mylog ("Mirrored %s\n", name);
      times.actime = mtime;
      times.modtime = mtime;
      utime (name, &times);
  
      /* do we try to extract file.txt ? */

      if (extract == 1)
      {
        strcpy (thisdirname, name);
        *(strrchr(thisdirname, '/')) = 0;
  
        /* here is why we require dos merge .... it would be better to use
           UNIX programs (such as unzip, unarj etc).  However, unarj (the one
           I'm interested in) can't extract one file without modification */
  
        /* dos should work but not with the 30 day evaluation of merge
           3.2.2 */
        strcpy (basename, strrchr(name, '/')+1);
        if (strstr (basename, ".exe") != NULL)
        {
          *(strstr (basename, ".exe")) = 0;
          if (debug)
            sprintf(command, "/usr/bin/dos -p +b `mapfname %s` -sy %s.txt", name, basename);
          else
            sprintf(command, "/usr/bin/dos -p +b %s -sy %s.txt >/dev/null 2>&1", name, basename);
          mylog ("Attempting to unpack %s.txt from %s\n", basename, name);
          if (system (command) != -1)
          {
            char a[MAXNAMLEN], b[MAXNAMLEN];
            sprintf (a, "%s.txt", basename); 
            sprintf (b, "%s/%s.txt", thisdirname, basename); 
            rename (a, b);
          }
        }
  
        strcpy (basename, strrchr(name, '/')+1);
        if (strstr (basename, ".tar") != NULL)
        {
          *(strstr (basename, ".tar")) = 0;
          if (debug)
            sprintf(command, "/usr/bin/tar -xf %s %s.txt", name, basename);
          else
            sprintf(command, "/usr/bin/tar -xf %s %s.txt >/dev/null 2>&1", name, basename);
          mylog ("Attempting to unpack %s.txt from %s\n", basename, name);
          if (system (command) != -1)
          {
            char a[MAXNAMLEN], b[MAXNAMLEN];
            sprintf (a, "%s.txt", basename); 
            sprintf (b, "%s/%s.txt", thisdirname, basename); 
            rename (a, b);
          }
        }
      }
    }
  }
}

mydelete (char *name)
{
  DIR *dp;
  struct dirent *de;
  struct stat st;
  char fn[MAXNAMLEN];
  int l;

  /* recursivly delete the file name */

  if ( (strcmp (name, ".") == 0) || (strcmp (name, "..") == 0) )
    return;

  l=0;
  if (stat(name,&st)==0)
    l=1;
  if (lstat(name,&st)==0)
    l=1;
  if (l==0)
    mylog ( "delete: cannot stat %s\n", name);
  else
  {
    if ( (st.st_mode & S_IFMT) == S_IFDIR)
    {
      if ( (dp=opendir(name)) == NULL )
        mylog ("Cannot open local directory (delete) %s\n", name);
      else
      {
        while ( (de = readdir(dp)) != NULL )
        {
          if ( (strcmp (de->d_name, ".") != 0) && (strcmp (de->d_name, "..") != 0) )
          {
            sprintf(fn,"%s/%s",name,de->d_name);
            mydelete (fn);
          }
        }
        closedir(dp);
      }
      mylog ("Removing local directory %s\n", name);
      rmdir (name);
    }
    else
    {
      mylog ("Deleting local file %s\n", name);
      unlink (name);
    }
  }
}

mymkdir (char *name)
{
  /* make a local directory */

  mylog ("mkdir %s\n", name);
  if (mkdir (name, S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH) != 0)
    mylog ("Can't mkdir %s\n", name);
}

myftpdebug (FTP *ftp, int code, char *msg)
{
  if (code < 0)
  {
    mylog ("FTP debug code %d (%s)\n", code, msg);
  }
}


myftperror (FTP *ftp, int code, char *msg)
{
  int c,d;

  if (code < 0)
  {
    switch (code)
    {
      case -6: /* too many open files .... we are in trouble! */
                 mylog ("FTP error code %d (%s)\n", code, msg);
                 myquit ("Too many open files .... exiting\n");
                 break;
      case -221: /* the end .. */
                 mylog ("Mirror finished\n");
                 exit (0);
                 break;
      case -550: /* permission denied - shouldn't happen as perms are now checked*/
                 FtpAbort (ftp);
                 break;
      default:
                 if (debug)
                 {
                   mylog ("FTP error code %d (%s)\n", code, msg);
                   mylog ("FTP Error occured, loging out\n");
                 }
                 FtpQuickBye (ftp);
                 sleep (10);
                 /* it seems the ftp libs can keep files open! */
                 for (c=30; c>2; c--)
                   close(c);
                 if (debug)
                   mylog ("Re-trying connection\n");
                 d=0;
                 while (((c=FtpLogin(&ftp,host,user,pass,NULL))<0) && (d<LOGINTRIES))
                 {
                   sleep (LOGINDELAY);
                   d++;
                 }
                 if (c<0)
                   myquit ("Cannot login to %s (%d)\n", host,c);
                 if ((c=FtpChdir(ftp,pwd))<0)
                   myquit ("Cannot cd to %s (%d)\n", pwd, c);
                 if ((c=FtpBinary(ftp))<0)
                   myquit ("Cannot use binary mode (%d)\n",c);
                 if (debug)
                   mylog ("Continuing with copy\n");
                 longjmp (getstack,1);
                 break;
    }
  }
}

main(int argc, char **argv)
{
  int c,d;

  verbose = 0;
  debug = 0;
  list = 0;
  extract = 0;
  cmp = 0;

  while ((c = getopt (argc, argv, "lvdec")) != EOF)
  switch (c) 
  {
    case 'd':  debug=1;
               break;
    case 'v':  verbose=1;
               break;
    case 'l':  list=1;
               break;
    case 'e':  extract=1;
               break;
    case 'c':  cmp=1;
               break;
    case '?':  myquit(USAGE);
               break;
  }

  maxfilesize=ulimit (UL_GFILLIM) * 512;
  sigignore (SIGPIPE);

  ftp = 0;
  
  strcpy (host, argv[optind+1]);
  strcpy (user, argv[optind+2]);
  strcpy (pass, argv[optind+3]);
  strcpy (pwd, argv[optind+4]);

  if ( (argc - optind) != 5)
    myquit(USAGE);

  /* read passwd on stdin ? */
  if (strcmp ("-", pass) == 0)
    fscanf (stdin, "%s", &pass);
  
  if (debug)
    FtplibDebug(yes);

  if (chdir (argv[optind])!=0)
    myquit ("Cannot cd to local directory %s\n", argv[optind]);

  d=0;
  while (((c=FtpLogin(&ftp,host,user,pass,NULL))<0) && (d<LOGINTRIES))
  {
    sleep (LOGINDELAY);
    d++;
  }
  if (c<0)
    myquit ("Cannot login to %s (%d)\n", host,c);
  if ((c=FtpChdir(ftp,pwd))<0)
    myquit ("Cannot cd to %s (%d)\n", pwd, c);
  if ((c=FtpBinary(ftp))<0)
    myquit ("Cannot use binary mode (%d)\n",c);
  
  doit(".");
  FtpBye(ftp);
  exit(0);
}

doit(char *dirname)
{
  DIR *dp;
  FILE *fp;
  struct dirent *de;
  struct tm tm, *today;
  char n[256],fn[MAXNAMLEN];
  struct stat st;
  time_t k;
  
  int i, j, l, exists;

  int local_files, remote_files;
  char *local_names, *remote_names;
  char *local_link, *remote_link;
  off_t *local_size, *remote_size;
  time_t *local_mtime, *remote_mtime;
  mode_t *local_mode, *remote_mode;

  char mode[MAXNAMLEN], temp[MAXNAMLEN], month[MAXNAMLEN], time_year [MAXNAMLEN], name[MAXNAMLEN], link[MAXNAMLEN], username[MAXNAMLEN], tempname[MAXNAMLEN];

  char tempfile[MAXNAMLEN], newname[MAXNAMLEN],basename[MAXNAMLEN];

  int date, size;

  FtpSetErrorHandler(ftp,myftperror);
  FtpSetIOHandler(ftp,myftperror);
  FtpSetDebugHandler(ftp,myftpdebug);

  fp = NULL;


/*
  FtpSetErrorHandler(&FtpInit,myftperror);
  FtpSetIOHandler(&FtpInit,myftperror);
  FtpSetDebugHandler(&FtpInit,myftpdebug);
*/

  tmpnam(tempfile);

  setjmp (getstack);

  if (fp != NULL)
    fclose (fp);
  if (stat(tempfile,&st)==0)
    unlink (tempfile); /* just in case we are jumping around too much ... */

  sprintf (temp, "-al %s", dirname);
  FtpDirectory (ftp, temp, tempfile );
#ifdef DEBUGTEMP
  fprintf (stderr, "Creating %s\n", tempfile);
#endif

  /* What is today ..*/
  k = time((time_t)0);
  /* gmtime_r (&k, &today); */
  today = gmtime (&k);

  /* if we just want a listing, no need to check out local dir */

  if (list == 0)
  {
    /* first, count the number of local files */
    if ( (dp=opendir(dirname)) == NULL )
    {
      mylog ("Cannot open local directory (listing) %s\n", dirname);
#ifdef DEBUGTEMP
      fprintf (stderr, "removing %s\n", tempfile);
#endif
      unlink (tempfile);
      return;
    }
    local_files = 0;
    while ( (de = readdir(dp)) != NULL )
      local_files++;
    closedir(dp);

    /* malloc the arrays */
#ifdef DEBUGMALLOC
    fprintf (stderr, "mallocing local for %s\n", dirname);
#endif
    if ( (local_names = (char *) malloc ((local_files+1) * MAXNAMLEN)) == NULL)
      myquit ("Cannot malloc\n");
    if ( (local_size = (off_t *) malloc ((local_files+1) * sizeof(off_t))) == NULL)
      myquit ("Cannot malloc\n");
    if ( (local_mtime = (time_t *) malloc ((local_files+1) * sizeof(time_t))) == NULL)
      myquit ("Cannot malloc\n");
    if ( (local_mode = (mode_t *) malloc ((local_files+1) * sizeof(mode_t))) == NULL)
      myquit ("Cannot malloc\n");
    if ( (local_link = (char *) malloc ((local_files+1) * MAXNAMLEN)) == NULL)
      myquit ("Cannot malloc\n");
  
    /* now build up the local filenames */

    if ( (dp=opendir(dirname)) == NULL )
    {
      mylog ("Cannot open local directory (mirroring) %s\n", dirname);
#ifdef DEBUGTEMP
      fprintf (stderr, "Removing %s\n", tempfile);
#endif
      unlink (tempfile);
#ifdef DEBUGMALLOC
      fprintf (stderr, "Freeing local for %s\n", dirname);
#endif
      free (local_link);
      free (local_names);
      free (local_size);
      free (local_mtime);
      free (local_mode);
      return;
    }
    local_files = 0;
    while ( (de = readdir(dp)) != NULL )
    {
      if ( (strcmp (de -> d_name, ".") == 0) | (strcmp (de -> d_name, "..") == 0) )
        continue;
      sprintf(fn,"%s/%s",dirname,de->d_name);
      l = 0; /* stat flag */
      if (stat(fn,&st)==0)
        l=1;
      if (lstat(fn,&st)==0)
        l=1;
      if (l==0)
        mylog ( "cannot stat %s\n", fn);
      else
      {
        strncpy ( &local_names[local_files * MAXNAMLEN], de -> d_name, MAXNAMLEN);
        local_size[local_files] = st.st_size;
        local_mtime[local_files] = st.st_mtime;
        local_mode[local_files] = st.st_mode;
        if ((st.st_mode & S_IFMT) == S_IFLNK)
        {
          if ((l=readlink (fn, link, MAXNAMLEN)) == -1 )
            mylog ( "cannot readlink %s\n", fn);
          else
          { 
            link[l]=0;
            strncpy ( &local_link[local_files * MAXNAMLEN], link, MAXNAMLEN);
          }
        }

        local_files++;
      }
    }  
    closedir(dp);
  }

  /* okay, now onto the remote system */

  if ( (fp = fopen (tempfile, "r")) == NULL)
  {
    mylog ("Remote dir failed on %s\n", dirname);
#ifdef DEBUGTEMP
    fprintf (stderr, "Removing %s\n", tempfile);
#endif
    unlink (tempfile);
    if (list==0)
    {
#ifdef DEBUGMALLOC
      fprintf (stderr, "Freeing local for %s\n", dirname);
#endif
      free (local_link);
      free (local_names);
      free (local_size);
      free (local_mtime);
      free (local_mode);
    }
    return;
  }
  remote_files=0;
  fgets (temp, MAXNAMLEN, fp);
  while ( fgets (temp, MAXNAMLEN, fp) != NULL)
    remote_files++;
  fclose (fp);

  /* malloc the arrays */
#ifdef DEBUGMALLOC
  fprintf (stderr, "mallocing remote for %s\n", dirname);
#endif
  if ( (remote_names = (char *) malloc ((remote_files+1) * MAXNAMLEN)) == NULL)
    myquit ("Cannot malloc\n");
  if ( (remote_size = (off_t *) malloc ((remote_files+1) * sizeof(off_t))) == NULL)
    myquit ("Cannot malloc\n");
  if ( (remote_mtime = (time_t *) malloc ((remote_files+1) * sizeof(time_t))) == NULL)
    myquit ("Cannot malloc\n");
  if ( (remote_mode = (mode_t *) malloc ((remote_files+1) * sizeof(mode_t))) == NULL)
    myquit ("Cannot malloc\n");
  if ( (remote_link = (char *) malloc ((remote_files+1) * MAXNAMLEN)) == NULL)
    myquit ("Cannot malloc\n");

  if ( (fp = fopen (tempfile, "r")) == NULL)
  {
    mylog ("Remote dir failed on %s\n", dirname);
#ifdef DEBUGTEMP
    fprintf (stderr, "Removing %s\n", tempfile);
#endif
    unlink (tempfile);
    if (list==0)
    {
#ifdef DEBUGMALLOC
      fprintf (stderr, "Freeing local for %s\n", dirname);
#endif
      free (local_link);
      free (local_names);
      free (local_size);
      free (local_mtime);
      free (local_mode);
    }
#ifdef DEBUGMALLOC
    fprintf (stderr, "Freeing remote for %s\n", dirname);
#endif
    free (remote_link);
    free (remote_names);
    free (remote_size);
    free (remote_mtime);
    free (remote_mode);
    return;
  }
  remote_files=0;
  fgets (temp, MAXNAMLEN, fp);
  while ( (fgets (temp, MAXNAMLEN, fp) != NULL))
  {
    sscanf (temp, "%s %*s %s %*s %d %s %d %s %s %*s %s\n", &mode, &username, &size, &month, &date, &time_year, &name, &link);
    if (( (strcmp (user, username) != 0) & (mode[1] == 'r')) & (mode[7] != 'r'))
    {
      mylog ("Unable to read %s/%s - skipping\n",dirname, name);
      continue;
    }

    if ( (strcmp (name, ".") == 0) | (strcmp (name, "..") == 0) )
    {
      continue;
    }
    strncpy ( &remote_names[remote_files * MAXNAMLEN], name , MAXNAMLEN);
    sprintf (temp, "%02d %s %s", date, &month, &time_year);
    if (strchr (temp, ':') != NULL)
    {
      if (strptime (temp, "%d %b %R", &tm ) == NULL)
        mylog ("strptime failed\n");
      /* we don't know the year ... */
      if (tm.tm_mon <= today->tm_mon)
        tm.tm_year = today->tm_year;
      else
        tm.tm_year = today->tm_year - 1;
    }
    else
    {
      if (strptime (temp, "%d %b %Y", &tm ) == NULL)
        mylog ("strptime failed\n");
      /* this info is unknown */
      tm.tm_min = 0;
      tm.tm_hour = 0;
    }
    tm.tm_sec = 0;

    if (mode[0] == 'd')
      remote_mode[remote_files] = S_IFDIR;
    else
      if (mode[0] == 'l')
      {
        remote_mode[remote_files] = S_IFLNK;
        strncpy ( &remote_link[remote_files * MAXNAMLEN], link , MAXNAMLEN);
      }
      else
        if (mode[0] == '-')
          remote_mode[remote_files] = S_IFREG;
        else
          remote_mode[remote_files] = 0;

/*    mylog ( "name[%d]=%s, mode=%d\n", remote_files, &remote_names[remote_files * MAXNAMLEN], remote_mode[remote_files]); */
    remote_size[remote_files] = size;
    remote_mtime[remote_files] = mktime (&tm);
    remote_files++;
  }

  fclose (fp);
#ifdef DEBUGTEMP
  fprintf (stderr, "Removing %s\n", tempfile);
#endif
  unlink (tempfile);

  if (list)
  {
    /* do a simple list */

    for (i=0; i<remote_files;i++)
    {
      if ((remote_mode[i] & S_IFMT) == S_IFREG)
        printf ("%s/%s %ld\n",dirname, & remote_names[i * MAXNAMLEN], remote_size[i]);
      if ((remote_mode[i] & S_IFMT) == S_IFDIR)
      {
        sprintf (temp, "%s/%s", dirname, &remote_names[i * MAXNAMLEN]);
        printf ("%s/\n", temp);
        /* decend into this dir */
        doit (temp);
      }
      if ((remote_mode[i] & S_IFMT) == S_IFLNK)
        printf ("%s/%s -> %s\n",dirname, & remote_names[i * MAXNAMLEN], &remote_link[i * MAXNAMLEN]);
    }
  }
  else
  {

    /* okay, we can now work out what needs to be done */

    /* i)    If a filename in the remote list doesn't exist in the local, then
             copy the file */
    /* ii)   If the file name exists in both lists AND the size or timestamp
             differs, then copy the file */

    for (i=0; i<remote_files;i++)
    {
      if ((remote_mode[i] & S_IFMT) == S_IFREG)
      {
        exists = 0;
        for (j=0; j<local_files; j++)
        {
          if ((strcmp(&remote_names[i * MAXNAMLEN], &local_names[j * MAXNAMLEN]) == 0))
          {
            exists = 1;
            if (remote_size[i] != local_size[j])
            {
              sprintf (tempname, "%s/%s", dirname, &remote_names[i * MAXNAMLEN]);
              /* mylog ("Need to mirror %s (size difference)\n", tempname); */
              /* we are wasting a local disk access here ... should put the
                 file mode into an array ... but life is short!" */
              if (chmod ( tempname, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)==0)
                mycopy ( tempname, remote_mtime[i], remote_size[i]);
              else
                mylog ("Cannot write to local file %s\n", tempname);
            }
            else
              if (remote_mtime[i] != local_mtime[j])
              {
                sprintf (tempname, "%s/%s", dirname, &remote_names[i * MAXNAMLEN]);
                /* mylog ("Need to mirror %s (mtime difference)\n", tempname); */
                if (chmod ( tempname, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)==0)
                  mycopy (tempname, remote_mtime[i], remote_size[i]);
                else
                  mylog ("Cannot write to local file %s\n", tempname);
              }
          }
        }
        if (exists == 0)
        {
          sprintf (tempname, "%s/%s", dirname, &remote_names[i * MAXNAMLEN]);
          /* mylog ("Need to mirror %s (doesn't exist)\n", tempname); */
          mycopy (tempname, remote_mtime[i], remote_size[i]);
        }
      }
  
    /* a)    If the remote filename is a directory AND the local directory
             exists then doit (file) */
    /* b)    If the remote filename is a directory AND the local directory
             doesn't exist then mkdir (file); then doit (file)  */
      else
      {
        if ((remote_mode[i] & S_IFMT) == S_IFDIR)
        {
          exists = 0;
          for (j=0; j<local_files; j++)
            if ((strcmp(&remote_names[i * MAXNAMLEN], &local_names[j * MAXNAMLEN]) == 0))
              exists = 1;
          if (exists == 0)
          {
            sprintf (temp, "%s/%s", dirname, &remote_names[i * MAXNAMLEN]);
            mymkdir (temp);
          }
          /* decend into this dir */
          sprintf (temp, "%s/%s", dirname, &remote_names[i * MAXNAMLEN]);
          doit (temp);
        }
        else
        {
          if ((remote_mode[i] & S_IFMT) == S_IFLNK)
          {
            sprintf (temp, "%s/%s", dirname, &remote_names[i * MAXNAMLEN]);
            exists=-1;
            for (j=0; j<local_files; j++)
            {
              if ((strcmp(&remote_names[i * MAXNAMLEN], &local_names[j * MAXNAMLEN]) == 0))
                exists=j;
            }
            if (exists != -1)
            {
              if ((strcmp(&remote_link[i * MAXNAMLEN], &local_link[exists * MAXNAMLEN]) != 0))
              {
                mydelete (temp);
                mylog ("Re-creating symbolic link %s -> %s\n", temp, &remote_link[i*MAXNAMLEN]);
                symlink (&remote_link[i*MAXNAMLEN], temp);
              }
            }
            else
            {
              mylog ("Creating symbolic link %s -> %s\n", temp, &remote_link[i*MAXNAMLEN]);
              symlink (&remote_link[i*MAXNAMLEN], temp);
            }
          }
        }
      }
    }

    /* iii)  If the file exists in the local list but doesn't on the remote,
             then delete - unless you have a fred.txt with a fred.tar/fred.exe */
    for (j=0; j<local_files;j++)
    {
      exists = 0;
      for (i=0; i<remote_files; i++)
      {
        if ((strcmp(&remote_names[i * MAXNAMLEN], &local_names[j * MAXNAMLEN]) == 0))
          exists = 1;
      }
      if (exists == 0)
      {
        sprintf (temp, "%s/%s", dirname, &local_names[j * MAXNAMLEN]);
        strcpy (basename, strrchr(temp, '/')+1);
        /* check to see if we have a matching fred.txt with a .exe/.tar */
        if ((strstr (basename, ".txt") != NULL) && (extract==0))
        {
          exists = 0;
          strcpy (newname, basename);
          *(strstr(newname, ".txt"))=0;
          strcat (newname, ".exe");
          for (i=0; i<remote_files; i++)
          {
            if ((strcmp(&remote_names[i * MAXNAMLEN], newname) == 0))
            {
              mylog ("%s exists so %s isn't being deleted\n", newname, temp);
              exists = 1;
            }
          }
          strcpy (newname, basename);
          *(strstr(newname, ".txt"))=0;
          strcat (newname, ".tar");
          for (i=0; i<remote_files; i++)
          {
            if ((strcmp(&remote_names[i * MAXNAMLEN], newname) == 0))
            {
              mylog ("%s exists so %s isn't being deleted\n", newname, temp);
              exists = 1;
            }
          }
          if (exists==0)
            mydelete (temp);
        }
        else
          mydelete (temp);
      }
    }
  
    /* If a file is copied then it must be "touched" to make sure mtime is
     correct. */
  
  }

#ifdef DEBUGMALLOC
  fprintf (stderr, "Freeing remote for %s\n", dirname);
#endif
  free (remote_link);
  free (remote_names);
  free (remote_size);
  free (remote_mtime);
  free (remote_mode);
  
  if (list==0)
  {
#ifdef DEBUGMALLOC
    fprintf (stderr, "Freeing local for %s\n", dirname);
#endif
    free (local_link);
    free (local_names);
    free (local_size);
    free (local_mtime);
    free (local_mode);
  }

  return (0);
}
