/* Mirror directrory structure to another host                   */
/*                                                               */
/* Usage: mirror [-v][-d][-l] local_dir host user passwd remote_dir */ 
/*                                                               */
/*    -v verbose                                                 */
/*    -d debug                                                   */
/*    -l only list, don't copy                                   */
/*                                                               */
/* Performs a mirror of the remote directory to the local using  */
/* the ftp protocol.  Sub directories are also mirrored.  Files  */
/* and directories are deleted on the local machine when required*/
/*                                                               */
/* Mirroring is based on timestamp and file size info            */
/*                                                               */
/* ideas for future enhancements :-                              */
/*                                                               */
/*   Read full perms of remote machine to mirror perms as well   */
/*   Read sym links and create locally                           */
/*   More efficient backup by moving files instead of deleting   */
/*     and then check the list of "moved" files before every get */
/*     (really del the temp directory before exiting)            */
/*   Delete hard linked files                                    */
/*   Ability to mirror from more than one machine into one dir   */
/*   Auto re-try connection & re-connect if disconnected         */
/*     This will involve avoiding cd'ing on the remote machine   */
/*     and making use of an Ftp_reopen routine called from the   */
/*     my_error routine (called from FtpSetErrorHandler &        */
/*     FtpSetIOHandler)                                          */
/*   Allow the existance (don't delete) & extraction of readme's */
/*     eg tar -xf file.tar file.txt                              */
/*     eg dos -p +b file.exe -sy file.txt                        */
/*   Improve the code structure !!!                              */
/*   Clean up temp files on abnormal quit                        */
/*                                                               */
/*                                                               */
/*                                                               */
/*   Peter Lord (plord@novell.co.uk)                             */
/*                                                               */
/*   (C) Novell 1995                                             */
/*                                                               */
/*   Based on an "uftp" - uses libftp                            */
/*                                                               */


#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/dir.h>
#include <sys/stat.h>
#include <syslog.h>
#include <dirent.h>
#include <string.h>
#include <varargs.h>
#include <utime.h>
#include <unistd.h>
#include <FtpLibrary.h>

FTP *ftp;
int verbose, debug, list;

/* #define DEBUGMALLOC */
/* #define DEBUGTEMP */

#define USAGE "Usage: mirror [-v][-d][-l] local_dir host user passwd remote_dir\n\n	-v verbose\n	-d debug\n	-l only list, don't copy\n"

void mylog (format, va_alist) const char *format; va_dcl
{
  if (verbose)
  {
    /* log handler */
    va_list ap;
    va_start(ap);
    fprintf (stderr, "mirror: ");
    vfprintf(stderr, format, ap);
    va_end(ap);
  }
}

void myerror (format, va_alist) const char *format; va_dcl
{
  /* non-quit error handler */
  va_list ap;
  va_start(ap);
  fprintf (stderr, "mirror: Error: ");
  vfprintf(stderr, format, ap);
  va_end(ap);
}

void myquit (format, va_alist) const char *format; va_dcl
{
  /* quit handler */
  va_list ap;
  va_start(ap);
  fprintf (stderr, "mirror: Error: ");
  vfprintf(stderr, format, ap);
  va_end(ap);
  if (ftp > (FTP *)0)
    FtpBye(ftp);
  exit(1);
}

mycopy (char *from, char *to, time_t mtime)
{
  /* copy a file and "touch" it */
  int a;
  struct utimbuf times;

  a = FtpGet (ftp, from, to);

  if (a < 0)
  {
    myerror ("Cannot mirror %s\n", to);
    unlink (to);
  }
  else
  {
    times.actime = mtime;
    times.modtime = mtime;
    utime (to, &times);
  }
}

mydelete (char *name)
{
  /* delete the file name */

  /* note we should check to see if name is a directory and recursivily zap */

  unlink (name);
}

mymkdir (char *name)
{
  /* make a local directory */

  mylog ("mkdir %s (doesn't exist)\n", name);
  if (mkdir (name, S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH) != 0)
    mylog ("Can't mkdir %s\n", name);
}

main(int argc, char **argv)
{

  int c;

  verbose = 0;
  debug = 0;
  list = 0;

  while ((c = getopt (argc, argv, "lvd")) != EOF)
  switch (c) 
  {
    case 'd':  debug=1;
               break;
    case 'v':  verbose=1;
               break;
    case 'l':  list=1;
               break;
    case '?':  myquit(USAGE);
               break;
  }

  ftp = 0;

  if ( (argc - optind) != 5)
    myquit(USAGE);

  if (debug)
    FtplibDebug(yes);

  FtpLogin(&ftp,argv[optind+1],argv[optind+2],argv[optind+3],NULL);
  FtpChdir(ftp,argv[optind+4]);
  FtpBinary(ftp);
  doit(argv[optind]);
  FtpBye(ftp);
  exit(0);
}

doit(char *dirname)
{
  DIR *dp;
  FILE *fp;
  struct dirent *de;
  struct tm tm, today;
  char n[256],fn[MAXNAMLEN];
  struct stat st;
  time_t k;
  
  int i, j, exists;

  int local_files, remote_files;
  char *local_names, *remote_names;
  off_t *local_size, *remote_size;
  time_t *local_mtime, *remote_mtime;
  mode_t *local_mode, *remote_mode;

  char mode[MAXNAMLEN], temp[MAXNAMLEN], month[MAXNAMLEN], time_year [MAXNAMLEN], name[MAXNAMLEN];

  char tempfile[MAXNAMLEN];

  int date, size;

  tmpnam(tempfile);
  FtpDir (ftp, tempfile );
#ifdef DEBUGTEMP
  fprintf (stderr, "Creating %s\n", tempfile);
#endif

  /* What is today ..*/
  k = time((time_t)0);
  gmtime_r (&k, &today);

  /* if we just want a listing, no need to check out local dir */

  if (list == 0)
  {
    /* first, count the number of local files */
    if ( (dp=opendir(dirname)) == NULL )
    {
      mylog ("Cannot open local directory %s\n", dirname);
#ifdef DEBUGTEMP
      fprintf (stderr, "removing %s\n", tempfile);
#endif
      unlink (tempfile);
      return;
    }
    local_files = 0;
    while ( (de = readdir(dp)) != NULL )
      local_files++;
    closedir(dp);

    /* malloc the arrays */
#ifdef DEBUGMALLOC
    fprintf (stderr, "mallocing local for %s\n", dirname);
#endif
    if ( (local_names = (char *) malloc ((local_files+1) * MAXNAMLEN)) == NULL)
      myquit ("Cannot malloc\n");
    if ( (local_size = (off_t *) malloc ((local_files+1) * sizeof(off_t))) == NULL)
      myquit ("Cannot malloc\n");
    if ( (local_mtime = (time_t *) malloc ((local_files+1) * sizeof(time_t))) == NULL)
      myquit ("Cannot malloc\n");
    if ( (local_mode = (mode_t *) malloc ((local_files+1) * sizeof(mode_t))) == NULL)
      myquit ("Cannot malloc\n");
  
    /* now build up the local filenames */

    if ( (dp=opendir(dirname)) == NULL )
    {
      mylog ("Cannot open local directory %s\n", dirname);
#ifdef DEBUGTEMP
      fprintf (stderr, "Removing %s\n", tempfile);
#endif
      unlink (tempfile);
#ifdef DEBUGMALLOC
      fprintf (stderr, "Freeing local for %s\n", dirname);
#endif
      free (local_names);
      free (local_size);
      free (local_mtime);
      free (local_mode);
      return;
    }
    local_files = 0;
    while ( (de = readdir(dp)) != NULL )
    {
      if ( (strcmp (de -> d_name, ".") == 0) | (strcmp (de -> d_name, "..") == 0) )
        continue;
      sprintf(fn,"%s/%s",dirname,de->d_name);
      if ( stat(fn,&st) != 0 ) 
        mylog ( "cannot stat %s\n", de -> d_name);
      else
      {
        strncpy ( &local_names[local_files * MAXNAMLEN], de -> d_name, MAXNAMLEN);
        local_size[local_files] = st.st_size;
        local_mtime[local_files] = st.st_mtime;
        local_mode[local_files] = st.st_mode;
        local_files++;
      }
    }  
    closedir(dp);
  }

  /* okay, now onto the remote system */

  if ( (fp = fopen (tempfile, "r")) == NULL)
  {
    mylog ("Remote dir failed on %s\n", dirname);
#ifdef DEBUGTEMP
    fprintf (stderr, "Removing %s\n", tempfile);
#endif
    unlink (tempfile);
    if (list==0)
    {
#ifdef DEBUGMALLOC
      fprintf (stderr, "Freeing local for %s\n", dirname);
#endif
      free (local_names);
      free (local_size);
      free (local_mtime);
      free (local_mode);
    }
    return;
  }
  remote_files=0;
  fgets (temp, MAXNAMLEN, fp);
  while ( fgets (temp, MAXNAMLEN, fp) != NULL)
    remote_files++;
  fclose (fp);

  /* malloc the arrays */
#ifdef DEBUGMALLOC
  fprintf (stderr, "mallocing remote for %s\n", dirname);
#endif
  if ( (remote_names = (char *) malloc ((remote_files+1) * MAXNAMLEN)) == NULL)
    myquit ("Cannot malloc\n");
  if ( (remote_size = (off_t *) malloc ((remote_files+1) * sizeof(off_t))) == NULL)
    myquit ("Cannot malloc\n");
  if ( (remote_mtime = (time_t *) malloc ((remote_files+1) * sizeof(time_t))) == NULL)
    myquit ("Cannot malloc\n");
  if ( (remote_mode = (mode_t *) malloc ((remote_files+1) * sizeof(mode_t))) == NULL)
    myquit ("Cannot malloc\n");

  if ( (fp = fopen (tempfile, "r")) == NULL)
  {
    mylog ("Remote dir failed on %s\n", dirname);
#ifdef DEBUGTEMP
    fprintf (stderr, "Removing %s\n", tempfile);
#endif
    unlink (tempfile);
    if (list==0)
    {
#ifdef DEBUGMALLOC
      fprintf (stderr, "Freeing local for %s\n", dirname);
#endif
      free (local_names);
      free (local_size);
      free (local_mtime);
      free (local_mode);
    }
#ifdef DEBUGMALLOC
    fprintf (stderr, "Freeing remote for %s\n", dirname);
#endif
    free (remote_names);
    free (remote_names);
    free (remote_size);
    free (remote_mtime);
    free (remote_mode);
    return;
  }
  remote_files=0;
  fgets (temp, MAXNAMLEN, fp);
  while ( (fgets (temp, MAXNAMLEN, fp) != NULL))
  {
    sscanf (temp, "%s %*s %*s %*s %d %s %d %s %s\n", &mode, &size, &month, &date, &time_year, &name);
    if ( (strcmp (name, ".") == 0) | (strcmp (name, "..") == 0) )
    {
      continue;
    }
    strncpy ( &remote_names[remote_files * MAXNAMLEN], name , MAXNAMLEN);
    sprintf (temp, "%02d %s %s", date, &month, &time_year);
    if (strchr (temp, ':') != NULL)
    {
      if (strptime (temp, "%d %b %R", &tm ) == NULL)
        mylog ("strptime failed\n");
      /* we don't know the year ... */
      if (tm.tm_mon <= today.tm_mon)
        tm.tm_year = today.tm_year;
      else
        tm.tm_year = today.tm_year - 1;
    }
    else
    {
      if (strptime (temp, "%d %b %Y", &tm ) == NULL)
        mylog ("strptime failed\n");
      /* this info is unknown */
      tm.tm_min = 0;
      tm.tm_hour = 0;
    }
    tm.tm_sec = 0;

    if (mode[0] == 'd')
      remote_mode[remote_files] = S_IFDIR;
    else
      if (mode[0] == 'l')
        remote_mode[remote_files] = S_IFLNK;
      else
        if (mode[0] == '-')
          remote_mode[remote_files] = S_IFREG;
        else
          remote_mode[remote_files] = 0;

    remote_size[remote_files] = size;
    remote_mtime[remote_files] = mktime (&tm);
    remote_files++;
  }

  if (list)
  {
    /* do a simple list */

    for (i=0; i<remote_files;i++)
    {
      if ((remote_mode[i] & S_IFMT) == S_IFREG)
        printf ("%s/%s %ld\n",dirname, & remote_names[i * MAXNAMLEN], remote_size[i]);
      if ((remote_mode[i] & S_IFMT) == S_IFDIR)
      {
        sprintf (temp, "%s/%s", dirname, &remote_names[i * MAXNAMLEN]);
        printf ("%s/\n", temp);
        /* decend into this dir */
        FtpChdir (ftp, &remote_names[i * MAXNAMLEN]);
        doit (temp);
        FtpChdir (ftp, "..");
      }
      if ((remote_mode[i] & S_IFMT) == S_IFLNK)
        printf ("%s/%s ->\n",dirname, & remote_names[i * MAXNAMLEN]);
    }
  }
  else
  {

    /* okay, we can now work out what needs to be done */

    /* i)    If a filename in the remote list doesn't exist in the local, then
             copy the file */
    /* ii)   If the file name exists in both lists AND the size or timestamp
             differs, then copy the file */

    for (i=0; i<remote_files;i++)
    {
      if ((remote_mode[i] & S_IFMT) == S_IFREG)
      {
        exists = 0;
        for (j=0; j<local_files; j++)
        {
          if ((strcmp(&remote_names[i * MAXNAMLEN], &local_names[j * MAXNAMLEN]) == 0))
          {
            exists = 1;
            if (remote_size[i] != local_size[j])
            {
              sprintf (temp, "%s/%s", dirname, &remote_names[i * MAXNAMLEN]);
              mylog ("Copy %s to %s (size difference)\n", &remote_names[i * MAXNAMLEN], temp);
              mycopy (&remote_names[i*MAXNAMLEN], temp, remote_mtime[i]);
            }
            else
              if (remote_mtime[i] != local_mtime[j])
              {
                sprintf (temp, "%s/%s", dirname, &remote_names[i * MAXNAMLEN]);
                mylog ("Copy %s to %s (mtime difference)\n", &remote_names[i * MAXNAMLEN], temp);
                mycopy (&remote_names[i * MAXNAMLEN], temp, remote_mtime[i]);
              }
          }
        }
        if (exists == 0)
        {
          sprintf (temp, "%s/%s", dirname, &remote_names[i * MAXNAMLEN]);
          mylog ("Copy %s to %s (doesn't exist)\n", &remote_names[i * MAXNAMLEN], temp);
          mycopy (&remote_names[i * MAXNAMLEN], temp, remote_mtime[i]);
        }
      }
  
    /* a)    If the remote filename is a directory AND the local directory
             exists then doit (file) */
    /* b)    If the remote filename is a directory AND the local directory
             doesn't exist then mkdir (file); then doit (file)  */
      else
      {
        if ((remote_mode[i] & S_IFMT) == S_IFDIR)
        {
          exists = 0;
          for (j=0; j<local_files; j++)
            if ((strcmp(&remote_names[i * MAXNAMLEN], &local_names[j * MAXNAMLEN]) == 0))
              exists = 1;
          if (exists == 0)
          {
            sprintf (temp, "%s/%s", dirname, &remote_names[i * MAXNAMLEN]);
            mymkdir (temp);
          }
          /* decend into this dir */
          FtpChdir (ftp, &remote_names[i * MAXNAMLEN]);
          sprintf (temp, "%s/%s", dirname, &remote_names[i * MAXNAMLEN]);
          doit (temp);
          FtpChdir (ftp, "..");
        }
        else
        {
          if ((remote_mode[i] & S_IFMT) == S_IFLNK)
          {
            mylog ("%s/%s is a symbolic link\n", dirname, &remote_names[i*MAXNAMLEN]);
          }
        }
      }
    }

    /* iii)  If the file exists in the local list but doesn't on the remote,
             then delete */
    for (j=0; j<local_files;j++)
    {
      exists = 0;
      for (i=0; i<remote_files; i++)
      if ((strcmp(&remote_names[i * MAXNAMLEN], &local_names[j * MAXNAMLEN]) == 0))
        exists = 1;
      if (exists == 0)
      {
        sprintf (temp, "%s/%s", dirname, &local_names[j * MAXNAMLEN]);
        mylog ("local file %s deleted\n", &temp);
        mydelete (temp);
      }
    }
  
    /* If a file is copied then it must be "touched" to make sure mtime is
     correct. */
  
  }

#ifdef DEBUGMALLOC
  fprintf (stderr, "Freeing remote for %s\n", dirname);
#endif
  free (remote_names);
  free (remote_size);
  free (remote_mtime);
  free (remote_mode);
  
  if (list==0)
  {
#ifdef DEBUGMALLOC
    fprintf (stderr, "Freeing local for %s\n", dirname);
#endif
    free (local_names);
    free (local_size);
    free (local_mtime);
    free (local_mode);
  }

#ifdef DEBUGTEMP
  fprintf (stderr, "Removing %s\n", tempfile);
#endif
  unlink (tempfile);

  return (0);
}
